﻿using UnityEngine;
using Windows.Kinect;
using System.Runtime.InteropServices;
using System;

public class CoordinateMapperManager : MonoBehaviour
{
	private DepthSpacePoint[] _DepthCoordinates;

    private KinectModule.SourceBuffer _SourceBuffer;
    private KinectModule.KinectAccessManager _KinectAccessManager;

	long frameCount = 0;

	double elapsedCounter = 0.0;
	double fps = 0.0;
	
	Texture2D _ColorRGBX;

	bool nullFrame = false;

    public Texture2D ColorTexture
    { get { return _ColorRGBX; } }
    public byte[] BodyIndexBuffer
    { get { return _SourceBuffer.BodyIndexBuffer; } }
    public DepthSpacePoint[] DepthCoordinates
    { get { return _DepthCoordinates; } }
    public Body[] BodyBuffer
    { get { return _SourceBuffer.BodyBuffer; } }

    void Awake()
	{
        _ColorRGBX = new Texture2D (KinectModule.SourceBuffer.ColorWidth, KinectModule.SourceBuffer.ColorHeight, TextureFormat.RGBA32, false);

		_DepthCoordinates = new DepthSpacePoint[KinectModule.SourceBuffer.ColorWidth * KinectModule.SourceBuffer.ColorHeight];

        _KinectAccessManager.Access();
    }

	void OnGUI () 
	{
		GUI.Box (new Rect(10, 10, 200, 30), "FPS: " + fps.ToString("0.00"));

		if (nullFrame)
			GUI.Box (new Rect(10, 50, 200, 30), "NULL MSFR Frame");
	}

	void ProcessFrame()
	{
		var DepthData = GCHandle.Alloc(_SourceBuffer.DepthBuffer, GCHandleType.Pinned);
		var pDepthCoordinatesData = GCHandle.Alloc(_DepthCoordinates, GCHandleType.Pinned);

		_KinectAccessManager.CoordinateMapper.MapColorFrameToDepthSpaceUsingIntPtr(
			DepthData.AddrOfPinnedObject(), 
			(uint)_SourceBuffer.DepthBuffer.Length * sizeof(ushort),
			pDepthCoordinatesData.AddrOfPinnedObject(), 
			(uint)_DepthCoordinates.Length);

		pDepthCoordinatesData.Free();
		DepthData.Free();

		_ColorRGBX.LoadRawTextureData(_SourceBuffer.ColorBuffer);
		_ColorRGBX.Apply();
	}

    void GetFPS()
    {
        elapsedCounter += Time.deltaTime;
        if (elapsedCounter > 1.0)
        {
            fps = frameCount / elapsedCounter;
            frameCount = 0;
            elapsedCounter = 0.0;
        }
        frameCount++;
    }
	
	void Update()
	{
        GetFPS();

		var sourceFrame = _KinectAccessManager.SourceFrame;
        if ((nullFrame = sourceFrame == null))
			return;

        _SourceBuffer.UpdateBuffers(sourceFrame);

        ProcessFrame();
	}

	void OnApplicationQuit()
	{
        _DepthCoordinates = null;

        _KinectAccessManager.Dispose();
	}
}

namespace KinectModule
{
    public class SourceBuffer
    {
        public const int DepthWidth = 512;
        public const int DepthHeight = 424;
        public const int ColorWidth = 1920;
        public const int ColorHeight = 1080;
        public const int BodyCount = 6;

        private byte[] _ColorBuffer = new byte[ColorWidth * ColorHeight * 4],
                       _BodyIndexBuffer = new byte[DepthWidth * DepthHeight];
        private ushort[] _DepthBuffer = new ushort[DepthWidth * DepthHeight];
        private Body[] _BodyBuffer = new Body[BodyCount];

        public byte[] ColorBuffer
        { get { return _ColorBuffer; } }
        public byte[] BodyIndexBuffer
        { get { return _BodyIndexBuffer; } }
        public ushort[] DepthBuffer
        { get { return _DepthBuffer; } }
        public Body[] BodyBuffer
        { get { return _BodyBuffer; } }

        public void UpdateBuffers(MultiSourceFrame sourceFrame)
        {
            UpdateColorBuffer(sourceFrame);
            UpdateBodyIndexBuffer(sourceFrame);
            UpdateDepthBuffer(sourceFrame);
            UpdateBodyBuffer(sourceFrame);
        }

        private void UpdateColorBuffer(MultiSourceFrame sourceFrame)
        {
            using (var colorFrame = sourceFrame.ColorFrameReference.AcquireFrame())
                if (colorFrame != null)
                {
                    var ColorData = GCHandle.Alloc(_ColorBuffer, GCHandleType.Pinned);
                    colorFrame.CopyConvertedFrameDataToIntPtr(ColorData.AddrOfPinnedObject(), (uint)_ColorBuffer.Length, ColorImageFormat.Rgba);
                    ColorData.Free();
                }
        }

        private void UpdateBodyIndexBuffer(MultiSourceFrame sourceFrame)
        {
            using (var bodyIndexFrame = sourceFrame.BodyIndexFrameReference.AcquireFrame())
                if (bodyIndexFrame != null)
                {
                    var BodyIndexData = GCHandle.Alloc(_BodyIndexBuffer, GCHandleType.Pinned);
                    bodyIndexFrame.CopyFrameDataToIntPtr(BodyIndexData.AddrOfPinnedObject(), (uint)_BodyIndexBuffer.Length);
                    BodyIndexData.Free();
                }
        }

        private void UpdateDepthBuffer(MultiSourceFrame sourceFrame)
        {
            using (var depthFrame = sourceFrame.DepthFrameReference.AcquireFrame())
                if (depthFrame != null)
                {
                    var DepthData = GCHandle.Alloc(_DepthBuffer, GCHandleType.Pinned);
                    depthFrame.CopyFrameDataToIntPtr(DepthData.AddrOfPinnedObject(), (uint)_DepthBuffer.Length * sizeof(ushort));
                    DepthData.Free();
                }
        }

        private void UpdateBodyBuffer(MultiSourceFrame sourceFrame)
        {
            using (var bodyFrame = sourceFrame.BodyFrameReference.AcquireFrame())
                if (bodyFrame != null)
                    bodyFrame.GetAndRefreshBodyData(_BodyBuffer);
        }
    }

    public class KinectAccessManager : IDisposable
    {
        private KinectSensor _Sensor;
        private CoordinateMapper _CoordinateMapper;
        private MultiSourceFrameReader _SourceFrameReader;

        private bool _IsSensorOpen = false;
        private bool _IsFrameReaderOpen = false;

        public bool IsAccess
        { get { return _IsSensorOpen && _IsFrameReaderOpen; } }
        public CoordinateMapper CoordinateMapper
        { get { return IsAccess ? _CoordinateMapper : null; } }
        public MultiSourceFrame SourceFrame
        { get { return IsAccess ? _SourceFrameReader.AcquireLatestFrame() : null; } }

        public void Access()
        {
            OpenSensor();
            OpenFrameReader();

            _CoordinateMapper = _Sensor.CoordinateMapper;
        }

        private void OpenSensor()
        {
            if (_IsSensorOpen)
                return;

            _Sensor = KinectSensor.GetDefault();
            if (_Sensor == null)
            {
                //TODO: Popup Modal Window
                Debug.LogError("No ready Kinect found!");
                return;
            }

            _Sensor.Open();
            if (!_Sensor.IsOpen)
            {
                //TODO: Popup Modal Window
                Debug.LogError("Kinect sensor has problem!");
                CloseSensor();
                return;
            }

            _IsSensorOpen = true;
        }

        private void OpenFrameReader()
        {
            if (_IsFrameReaderOpen)
                return;

            _SourceFrameReader = _Sensor.OpenMultiSourceFrameReader(
                FrameSourceTypes.Color |
                FrameSourceTypes.Depth |
                FrameSourceTypes.BodyIndex |
                FrameSourceTypes.Body);

            _IsSensorOpen = (_SourceFrameReader != null);
        }

        public void Dispose()
        {
            CloseFrameReader();
            CloseSensor();
        }

        private void CloseSensor()
        {
            if (_IsSensorOpen)
            {
                _Sensor.Close();
                _IsSensorOpen = false;
            }
        }

        private void CloseFrameReader()
        {
            if (_IsFrameReaderOpen)
            {
                _SourceFrameReader.Dispose();
                _IsFrameReaderOpen = false;
            }
        }
    }
}
